#
# The intention of this file is to collect all the routines that filerunner
# uses to navigate file systems in one location.
# This includes deciding and action on requests for three different file
# system accesses:
# a) normal access
# b) Xftp access
# c) access to file systems that are temporary mounts on the current system
#
# The functionality provided includes:
#
# cd
# pwd
# directory read, includeing fcontrol if needed.
#
# The decision to use mode a) or c) is based on noticing if the file system
# being accessed is a temp mount or not.  This includes auto mount systems
# as well as usb and other memory sticks.
#
# frPwd should filter /tmp_mnt stuff out of the path. 
# How well does that work? Not
# Really we just want to get the true path with out the links..

# for getdirlist:
# file mtime glob & lstat
# 
# setProtectedServer takes a file/dir name and, if it is on a temp mount point
#                    checks to see if a server is already set up for that
#                    mp and that it is alive. If not one is set up.
#                    Return is either the handle for the server or null if
#                    the file/dir is not on a temp mp
proc setProtectedServer {dir} {
  if {![info exist ::testP] || $::testP != 1} {
    return {}
  }
  global mpServers
  if {[set mp [isTempMountPt $dir]] == {}} {return {}}
  if {[info exists mpServers($mp)]} {
    set serverStat [::server::Status $mpServers($mp)]
    if {$serverStat == "UP"} {
      return $mpServers($mp)
    } 
  }
  set mpServers($mp) [::server::Build ]
  if {$mpServers($mp) != 0} {
    # server is up, Now we need to provision it.
    # simple things like 'pwd' and 'cd' are direct.
    # More complex things take some code...
    foreach fun {cdAndTestR collectDirList \
		     getFileInfo collNsortFilesR LnkFile \
		     GetStringFromMode GetUidGidString} {
      ::server::FunToServer $mpServers($mp) $fun
    }
  }
  return $mpServers($mp) 
}

# This code uses procs:
# cdAndTest
#
# it uses upvar directory and produces upvar dirlist
# it also modifies lastmtime to the mtime of the given pwd (parm 2)
# This code can return an error    
proc collectDirList {lastmtime instPwd showAll} {
  upvar 1 directory directory dirlist dirlist
  upvar 1 $lastmtime mylastmtime
  #cdAndTest $directory/
  set mylastmtime [file mtime $instPwd]
  #frputs "Before glob on [pwd] " 
  set r [catch {glob -nocomplain *} dirlist]
  #frputs "After glob  " r
  if {$r} {
    return -code 1 $dirlist
  }
  if { $showAll } {
    # for reasons unknown -type hidden does NOT return .file where .file
    # is a link.  No other option on windows...
    if {$::tcl_platform(platform) == "unix" } {
      set r [catch {glob -nocomplain .*} dirlist2]
    } else {
      set r [catch {glob -nocomplain -type hidden *} dirlist2]
    }
    #frputs "After glob hidden  " r
    if {$r} {
      return -code 1 $dirlist2
    }
    set dirlist [concat $dirlist2 $dirlist]
  }
}
proc GetUidGidString { file uid gid } {
  global glob users groups
  
  if { $::tcl_platform(platform) == "windows"} {
    return ""
  }
  if { ![info exists users($uid)] || ![info exists groups($gid)] } {
    # Need to work for it...
    if { [catch {file attributes $file} gu] != 0 } {
      # rare but happens with dead links...
      return ""
    }
    #set gu [file attributes $file]
    set users($uid) [lindex $gu 3]
    set groups($gid) [lindex $gu 1]
  }
  return "$users($uid)/$groups($gid)" 
}
#
# getFileInfo call the following procs:
# LnkFile
# GetStringFromMode
# GetUidGidString
# The ladder two of the above use (set to linux)

# It uses upvars dirlist and directory as input
# and puts its results in upvars dl and sortlist
# parms: showDirs should be one of "dirsfirst, mixed or dirslast"
#        showSort should be one of "mtime, rmtime, atime, ratime,
#                                   ctime, rctime, size, extension or ?"
#                 (any value '?') will sort on name only
#
proc getFileInfo {showDirs showSort } {
  upvar dirlist dirlist directory directory dl dl sortlist sortlist

  set sortval_d [switch -exact $showDirs {
    dirsfirst {expr {0}}
    mixed     {expr {1}}
    dirslast  {expr {2}}
  }]

  foreach k $dirlist {
    #frputs "before Stat  " k
    if {[catch { file lstat "./$k" statinfo }] && \
	    [catch {file lstat $k statinfo}]} continue
    set filetype n

    if {($statinfo(mode) & 0170000) == 040000} {
      set filetype d
    }
    set linkname " "
    if {($statinfo(mode) & 0170000) == 0120000} {
      set filetype l
      catch {file readlink "./$k"} linkname
      if {[file isdirectory "./$k"]} {
	set filetype ld
      }
    } else {
      # check if it is a windows ".lnk" file
      if {[file extension "./$k"] == ".lnk"} {
	# For complete mount point protection we need to rewrite LnkFile...
	if {[catch {LnkFile ./$k to dir} out] == 0 && $out} {
	  set filetype [expr {$dir ? {wld} : {wl}}]
	  set linkname $to
	}	  
      }
    }
    set tmp [switch -exact $showSort {
      atime     {format "%011d" $statinfo(atime) }
      ctime     {format "%011d" $statinfo(ctime) }
      mtime     {format "%011d" $statinfo(mtime) }
      ratime    {format "%011d" [expr 2147483647 - $statinfo(atime)]}
      rctime    {format "%011d" [expr 2147483647 - $statinfo(ctime)]}
      rmtime    {format "%011d" [expr 2147483647 - $statinfo(mtime)]}
      size      {format "%011d" $statinfo(size) }
      extension {file extension $k}
      default   {expr { " "}}
    }]
    switch -exact $filetype {
      wl -
      n  -
      l  {set sortval 1$tmp}
      d   -
      wld -
      ld  {set sortval $sortval_d$tmp}
    }
    lappend dl [list $sortval $k $filetype $statinfo(size) \
		    $statinfo(mtime) [GetStringFromMode $statinfo(mode)] \
		    [GetUidGidString $directory/$k \
			 $statinfo(uid) $statinfo(gid)] \
		    $linkname $statinfo(nlink)\
		    $statinfo(atime) $statinfo(ctime)]
    lappend sortlist "$sortval $k"
  }
}

proc frPwd { } {
  global glob
  if { $::tcl_platform(platform) == "windows" } {
    return [pwd]
  }
  return [file dirname [file normalize [pwd]/x]]
}
proc NonLocalDir { dir } {
  # return 1 if not a local file system, else 0
  return  [catch {exec df -l $dir} ]
}
proc TryMakeNewDir { newdir } {
  set r [regexp {[a-z]?ftp://([^/]*)(.*)}\
		 $newdir match ftpI dir]
      if {$r} {
        Try { FTP_MkDir $ftpI "$dir" } "" 1
      } else {
        Try { file mkdir $newdir } "" 1
      }
}
# This code plants "cdAndTout" in the callers space!
# This is the first place we will see a temp mount point
# we provide an result that may be used by newpwd if the
# location is to become one of the panes.
proc cdAndTest {newpwd} {
  upvar cdAndTout cdAndTout mp mp
  set mp [setProtectedServer $newpwd]
  frputs "cdAndTest  " mp
  if {$mp != {}} {
    lassign [::server::Call $mp cdAndTestR $newpwd] val cdAndTout
    frputs "cdAndTest rtn  " val cdAndTout
    if {$val == 0} {cd $newpwd}
    return $val
  }
  set cdAndTout "Not readable"
  if {![file readable $newpwd]} {return 1}
  set r [catch {cd $newpwd} cdAndTout]
  return $r
}
# Here is the server version of cdAndTest
proc cdAndTestR {newpwd} {
  set cdAndTout "Not readable"
  if {![file readable $newpwd]} {puts "err";return 1} ; # include remote tests
  set r [catch {cd $newpwd} cdAndTout]
  puts "[list $r $cdAndTout]"
  return
}


# I don't see how, short of doing all this locally to prevent 
# wandering into a temp mount space.
#
proc GetNewDir {inst } {
  global glob ignor_error_flag
  set info  [_ "File runner choose directory"]
  set r 1
  while {$r != 0} {
    set ignor_error_flag [_ "Can not go there (permission?)"]
    set r [catch {tk_chooseDirectory -title $info \
		      -initialdir [$glob(win,$inst).entry_dir get] \
		      -mustexist 0} newdir]
    unset ignor_error_flag
  }
  if {[file isdirectory $newdir]} {
    return $newdir
  }
  return {}
}

proc NewPwd {inst newpwd {ask 0} } { 
  global glob config

  set curdir $glob($inst,pwd)
  set info ""
  set tmp2 [string range $glob(${inst},newpwd_oldpwd) 0 5]
  set rqpwd $newpwd
  
  while { 1 } {
    if { ! [IsFTP $newpwd] } {
      # for reasons unknown file normalize will not remove the 
      # extra '/' in //foo, unless we add the '//.' at the end.
      set newpwd [file normalize  $newpwd//.]
    }
    if {  [IsFTP $newpwd] &&  ! $ask } {
      set mode ftp
      # The following returns continue or break as needed
      # It uses 'upvar' to set 'newpwd' with the desired value

      set newpwd [FtpCheckSyntax $inst $newpwd ]

      # End of ftp tests

    } else {
      if {$::tcl_platform(platform) == "windows"} {
	switch -regexp -nocase $rqpwd {
	  {^[a-z]:/\.\.$} {
	    set glob($inst,pwd) "<volume>"
	    set glob($inst,pwdTail) "<volume>"
	    return
	  }
	  {^<volume>/\.\.} {
	    set ask 1
	  }
	  {^<volume>/.*} {
	    set newpwd [string range $rqpwd 9 end]
	  }
	}
      }
      set mode normal
      if { $ask ||  $newpwd == "" &&  $curdir == "" } {
	set newpwd [GetNewDir $inst]
	if { $newpwd == "" && $curdir != "" } {
	  return
	}
      } else {
	# Its not an 'ask' so if the new is nil, just leave it all alone
	if { $newpwd == "" } {
	  return
	}
      }
      #puts $newpwd
      # if newpwd is ?:/.. or /.. or //*/.. set the ask flag and go round again
      if {[regexp -nocase -lineanchor \
	       {^[a-z]:/\.\.$|^/\.\.$|^//[^/]+/[^/]+/\.\.$} \
	       $newpwd] == 1} {
	# Yep, lets ask via the menu
	set ask 1
	continue
      }
      frputs "about to call cdAndTest  " newpwd
      if {[cdAndTest $newpwd] != 0} {
	frputs "bad rtn from cdAndTest try link  " newpwd cdAndTout
	set r [catch {file link $newpwd} tryL]
	if {$r == 0 && [file readable $tryL]} {
	  frputs "trying link  " tryL
	  set newpwd $tryL
	  continue
	}
	#	     puts "failed read test"
        # Failed to be readable or cd able (or both). trying to get
	# up the tree from the top??
        # Otherwise, we have a problem...
        set newpwd [simple_smart_dialog "."  \
			[_ "Error in path (not readable)"]\
			[_ "Error: %s\nPlease edit new path or cancel.\
                          OK or Return will create it." $cdAndTout] \
			$newpwd ]
	# If s/he returned cancel or abort or nil, if there 
	# is an old path just return, else ... well insist...
        if {$newpwd == "" && $curdir != ""} {return ""}
	TryMakeNewDir $newpwd
	set ask 0
        continue
      }
      # The following code does the pwd to a newly cd'ed dir except for
      # well darn, must be some file names that don't pass, but I forget.
      # We DO check <char>:*, /*, //<char><char>/* What is left?
      if {$config(cd_pwd)  &&
	  [regexp -nocase -lineanchor \
	       {^[a-z]:/.*$|^/.*$|^//[^/]+/[^/]+/.*$} $newpwd] == 1} {
	set r [catch {frPwd} out]
	if {$r} { 
	  PopError [_ "Trying to get directory info: %s" $out]
	  if { $curdir != "" } {return ""}
	  continue
	}
	set newpwd $out
      } else {
	# Evaluate xxx/yyy/zzz/../.. to xxx
	set newpwd [file normalize $newpwd]
      }
      # Here is where we set the new dir in glob(inst,pwd)
      # We also want to use ?: rather than ?:/ so we do that here
      regsub -nocase {(^[a-z]:)/$} $newpwd {\1} glob(${inst},pwd)
      set glob($inst,pwdTail) [file tail $glob($inst,pwd)]
      if {$glob($inst,pwdTail) == {} } {
	set glob($inst,pwdTail) "/"
      }
      set glob($inst,isTempFS) $mp
      break
    }
  }
  # End of while
  
  IsFTP $tmp2

  set glob(${inst},newpwd_oldpwd) $glob(${inst},pwd)

  AppendToDirHistory $glob(${inst},pwd)
  # Moving to a new dir, clear watch on old if needed
  ClearWatch $inst $glob(${inst},pwd)
  set oldy 0
  catch {set oldy [$glob(listbox,$inst).file index active]}
  set glob($inst,dirstack) \
      [linsert $glob($inst,dirstack) 0 [list $curdir $oldy]]
  if { [llength $glob($inst,dirstack)] > 110 } {
    set glob($inst,dirstack) [lrange $glob($inst,dirstack) 0 100]
  }
}

proc formatDF {free} {
  regsub -all {[^0-9]} $free "" free
  # should now be a pure number.  If nil, use zero...
  # better than failing in the following code...
  if { $free == "" } {
    set free 0
  }
  set indx 0
  while { $free > 1024. } {
    incr indx
    set free [expr {$free / 1024.}]
  }
  #  puts "$indx $free"
  set suf { " " K M G T}
  return [format {%.3f%s} $free [lindex $suf $indx]]
}

# This routine is to get the full disc size. On
# *nix system it also returns the mount point of the
# containing file system

proc GetDF { dir } {
  # Makes a stab at getting the full size of the disk
  global glob
  frputs "in getdf " dir [list exec cmd /C $dir]
  set mp {}
  if {$::tcl_platform(platform) != "windows"} {
    set r [catch {exec df $dir} out]
    set oout [split $out]
    set indx [lsearch -regexp $oout {[0-9]%}]
    if { $indx != -1 } {
      set mp [lindex $oout $indx+1]
      incr indx -1
      while {[set free [lindex $oout $indx]] == "{}" } {
	incr indx -1
      }
    } else {
      set free ""
    }
  } else {
    set cmd [list exec cmd /C dir]
    frputs "getdf windows " cmd
    set r [catch $cmd out]
    frputs "getdf windows result " r out
    if { $r } {
      set oout [split $out "\n"]
      set free [split [lindex $oout end]]
      set indx [lsearch -exact $free bytes]
      set free [lindex $free $indx-1]
      set indx 0
    } else {
      LogSilent $out
      return "??"
    }
  }
  return [list [formatDF $free] $mp]
}

# This routine (GetDirList) returns a sorted list of files in the current
# working dir
proc GetDirList { inst } {
  global config glob
  global ftp
  set directory $glob($inst,pwd)
  set result {}
  set sortlist {}
  set dl {}
  set type ""

  if { [IsFTP $directory] } {
    GetFtpDirList $ftpI $directory sortlist

  } else {
    # It is NOT ftp, following is normal local file processing.
    # need the '/' below to prevent misbehavior with 'c:' which, for some
    # reason is not the same as 'c:'.  
    # The '/' is ignored in other cases (we hope).
    set glob($inst,lasttime) [clock seconds]
    if {$directory == "<volume>"} {
      # This is only possible with windows...
      set dirlist {}
      foreach dir [file vol] {
	if {[file exists $dir]} {
	  if {[string length $dir] > 1 && [string index $dir end] == "/"} {
	    set dir [string range $dir 0 end-1]
	  }
	  lappend dirlist $dir
	}
      }
      getFileInfo $config(fileshow,dirs) $config(fileshow,sort) 
    } else {
      #frputs "Before dir  " directory
      ClearCherryPicker $inst
      if { $glob($inst,isTempFS) != {}} {
	lassign [::server::Call $glob($inst,isTempFS) \
		     collNsortFilesR $glob($inst,pwd) \
		     $config(fileshow,all) \
		     $config(fileshow,dirs) \
		     $config(fileshow,sort) ] val mtime dl sortlist
	if {$val != 0} {return -code 1 $mtime}
	frputs "Mtime update from to  " glob($inst,lastmtime) mtime
	set glob($inst,lastmtime) $mtime
      }
      if {$::tcl_platform(platform) == "windows" && $config(useDIR)} {
	getWindowsDIRfilelist  \
	    glob($inst,lastmtime) \
	    $glob($inst,pwd) \
	    $config(fileshow,all) \
	    $config(fileshow,dirs) \
	    $config(fileshow,sort)
	# windows code...
      } else {
	# *nix and tcl/windows
	cdAndTest $directory/
	collectDirList glob($inst,lastmtime) $glob($inst,pwd) \
	    $config(fileshow,all) 
	getFileInfo $config(fileshow,dirs) $config(fileshow,sort)
	set glob($inst,df) [GetDF $glob($inst,pwd)]
      }
    }

  }

  foreach el [lsort $config(sortoption) -indices $sortlist] {
    lappend result [lindex $dl $el]
  }
  return $result
}

# The following is the remote function to collect a dir list along
# with the file info. This code will run on the server and takes
# the current pwd and several config values. It returns sortlist and dl
# and lastmtime info
#
proc collNsortFilesR {instPwd showAll showDirs showSort} {
  set directory $instPwd
  cd $directory/
  puts stderr "Entered collNsortFilesR $instPwd $showAll $showDirs "
  set r [catch "collectDirList mtime $instPwd $showAll" err]
  if {$r != 0} {
    puts "$r $err $instPwd $showAll $::errorInfo"
    return
  }
  getFileInfo $showDirs $showSort
  puts "[list 0 $mtime $dl $sortlist]"
}

# Windows get dir info using windows DIR command

proc getWindowsDIRfilelist {lastmtime dir showall showDirs showSort} {
  global glob config  timeEnd eightX3Start eightX3End ownerEnd

  upvar 1 $lastmtime mylastmtime
  upvar dl dl sortlist sortlist
  upvar 1 inst inst
  
  cdAndTest $dir/
  set mylastmtime [file mtime $dir]
  set dirCmd "cmd /c dir/a /q /-c /x /a[expr {$showall ? {} : "-h"}]"
  set dirList {}
  set r [catch "exec {*}$dirCmd" dirList]
  if {$r != 0} {
    PopError $dirList
    set sortlist {}
    set dl {}
    return
  }
  set dirList [split $dirList \n]
  set glob($inst,df) [formatDF [lindex $dirList end end-2]]
  set indx -1
  # skip lines that don't start with something that looks date like
  while {$indx < [llength $dirList] && \
	     ![regexp {^[ ]*[0-9][0-9]?(/|-)} [lindex $dirList $indx]] } {
    incr indx
  }
  # This is the first line in the listing and should be for "."
  # This will be a <dir> line and will not have a short name
  # we use these characteristics to find the positions of
  # the various parts... We only need to do this once...
  if {![info exists ownerEnd]} {
    set fl [string trimright [lindex $dirList $indx]]
    if {[string index $fl end] != "."} {
      error "Failed to find reasonable return from 'cmd dir' on $dir/"
    }
    set ownerEnd [expr {[string length $fl] - 2}]
    set timeEnd [string first "<" $fl]
    while {[string index $fl [incr timeEnd -1]] == " " && $timeEnd > 0} {}
    set eightX3End [string first ">" $fl]
    while {[string index $fl [incr eightX3End]] == " " && \
	       $eightX3End < $ownerEnd} {}
    incr eightX3End -2
    set eightX3Start [expr {$eightX3End - 11}]
  }
  # Ok, now we go through the list and pluck out what is needed.
  # do the invariant stuff
  set sortval_d [switch -exact $showDirs {
    dirsfirst {expr {0}}
    mixed     {expr {1}}
    dirslast  {expr {2}}
  }]
  set needAtime [expr {$showSort == "atime" || \
			   $showSort == "ratime" || \
			   [lsearch -index 0 -exact \
				$config(ListBoxColumns,$inst) "atime"] != -1}]
  # set needCtime [expr {$showSort == "ctime" || \
  # 			   $showSort == "rctime" || \
  # 			   [lsearch -index 0 -exact \
  # 				$config(ListBoxColumns,$inst) "ctime"]}]
  set atime {}
  set ctime {}
  set regPat  {([^[]*) [[]([^]]*)}
  foreach ent [lrange $dirList $indx end] {
    #if {[string index $ent 0] == " "} {break}
    set mtim     [string range $ent 0 $timeEnd]
    set r [catch {clock scan $mtim} mtime]
    frputs "windows dir  " mtim mtime r
    if {$r != 0} {break}
    set sizeType  [string trim [string range $ent $timeEnd+1 $eightX3Start-1]]
    set shortName [string trim [string range $ent $eightX3Start $eightX3End]]
    set owner     [string trim [string range $ent $eightX3End+1 $ownerEnd]]
    set file      [string range $ent $ownerEnd+1 end]
    # that is all that we have, derive the rest
    # set mtime [clock scan $mtime]
    set size 0
    set filetype n
    set linkname {}
    switch -exact -nocase $sizeType {
      "<dir>"      {set filetype d}
      "<symlinkd>"  -
      "<junction>" {
	set filetype ld
	regexp $regPat $file n file linkname
      }
      "<symlink>" {
	set filetype ln
	regexp $regPat $file n file linkname
      }
      default    {
	set size $sizeType
 	if {[file extension $file] == ".lnk"} {
	  if {[catch {LnkFile ./$file to dir} out] == 0 && $out} {
	    set filetype [expr {$dir ? {wld} : {wl}}]
	    set linkname $to
	  }	  
	}

      }
    }
    if {$needAtime} {
      set atime [file atime $file]
    }
    # no ctime (hard access on windows)
    # if {$needCtime} {
    #   set ctime [file ctime $file]
    # }
    set tmp [switch -exact $showSort {
      atime     {format "%011d" $atime }
      ctime     {format "%011d" $ctime }
      mtime     {format "%011d" $mtime }
      ratime    {format "%011d" [expr 2147483647 - $atime]}
      rctime    {format "%011d" [expr 2147483647 - $ctime]}
      rmtime    {format "%011d" [expr 2147483647 - $mtime]}
      size      {format "%011d" $size }
      extension {file extension $file}
      default   {expr { " "}}
    }]
    switch -exact $filetype {
      wl -
      n  -
      l  {set sortval 1$tmp}
      d   -
      wld -
      ld  {set sortval $sortval_d$tmp}
    }
    lappend dl [list $sortval $file $filetype $size \
		    $mtime "mode" \
		    $owner \
		    $linkname \
		    $shortName\
		    $atime $ctime]
    lappend sortlist "$sortval $file"
  } ; # foreach
  
}
#
# All pure ftp stuff below here
#############################################################

proc GetFtpDirList {ftpI directory where} {
  global config glob ftp
  upvar $where sortlist inst inst dl dl
  #   regexp {[a-z]?ftp://([^/]*)(.*)} $directory match ftpI directory
  # cancel notify if to this pane
  if {$glob(notify,$inst) != "" } {
    if {$glob(notify,left) != $glob(notify,right) } {
      catch {$glob(notify,watchname) remove $glob(notify,$inst)}
    }
    set  glob(notify,$inst) ""
  }
  
  set dummy {{0 {Can't get file list, try again?} n 0 0 0 0 0}}
  frputs "Get dir  " ftpI directory
  set r [catch {FTP_CD $ftpI $directory} outp]
  if {$r != 0} {
    PopError $outp
    return $dummy
  }
  set r [catch {FTP_List $ftpI $config(fileshow,all)} dirlist]
  if {$r != 0} {
    PopError $dirlist
    return $dummy
  }

  # Example of output (now placed in dirlist)
  #total 3333 (optional)
  #drwxrwxr-x   8 root     wheel        1024 Mar 16 14:28 .
  #drwxrwxr-x   8 root     wheel        1024 Mar 16 14:28 ..
  #lrwxrwxrwx   1 root     root           11 Mar 16 14:28 apa -> welcome.msg
  #drwxrwxr-x   2 root     wheel        1024 Dec  3  1993 bin
  #drwxrwxr-x   2 root     wheel        1024 Aug 30  1993 etc
  #drwxrwxr-x   2 root     wheel        1024 Dec  3  1993 incoming
  #drwxrwxr-x   2 root     wheel        1024 Nov 17  1993 lib
  #drwxrwxr-x   3 root     wheel        1024 Mar 10 16:08 pub
  #drwxrwxr-x   3 root     wheel        1024 Aug 30  1993 usr
  #-rw-r--r--   1 root     root          312 Aug  1  1994 welcome.msg

  # from android... ls -l  (nasty nil size on l & d also messy dev entries)
  #lrwxrwxrwx root     root              2012-02-14 10:23 vendor -> /system/vendor
  #lrwxrwxrwx root     root              2012-02-14 10:23 d -> /sys/kernel/debug
  #lrwxrwxrwx root     root              2012-02-14 10:23 etc -> /system/etc
  #-rw-r--r-- root     root         3882 1969-12-31 17:00 ueventd.rc
  #-rw-r--r-- root     root          890 1969-12-31 17:00 ueventd.omap4430.rc
  #-rw-r--r-- root     root            0 1969-12-31 17:00 ueventd.goldfish.rc
  #drwxr-xr-x root     root              2012-02-10 20:38 system
  #drwxr-xr-x root     root              2012-02-14 10:23 sys
  #crw------- root     root       4,  67 2012-02-14 10:23 ttyS3
  #crw------- root     root       4,  66 2012-02-14 10:23 ttyS2

  #wuarchive.wustl.edu:
  #-rw-r--r--   1 0                      605 Sep 27 14:45 README.NFS
  #-rw-r--r--   1 0                      474 Sep 27 14:45 README.SIMTEL
  #lrwxrwxrwx   1 0                        9 Sep 26 12:56 bin -> ./usr/bin

  #ftp://reactor.actlab.com (Yucky WinNT output)
  #12-02-97  02:17AM       <DIR>          !Incoming
  #06-03-97  09:38PM       <DIR>          !support
  #06-03-97  09:38PM       <DIR>          7thlevel
  #06-03-97  09:38PM       <DIR>          access
  #06-03-97  09:38PM       <DIR>          accolade
  #06-03-97  09:39PM       <DIR>          Activision
  #09-11-96  07:10PM                 3592 ACTlogo.gif
  #06-03-97  09:40PM       <DIR>          Apogee
  #06-03-97  09:40PM       <DIR>          avalon
  #06-03-97  09:40PM       <DIR>          beam


  set sortval_d [switch -exact $config(fileshow,dirs) {
    dirsfirst {expr {0}}
    mixed     {expr {1}}
    dirslast  {expr {2}}
  }]
  set linkfound 0
  foreach k $dirlist {
    if { $k == "" } continue
    if { [string range $k 0 4] == "total" } continue

    set filetype fn
    set link {}
    set nlinks {}
    # Try regular parsing

    frputs "$k "

    set r [regexp {^([^ ])([^ ]+) *([0-9]+) +([^ ]+) +([^ ]+) +([0-9]+)\
		       +(............) (((.+) -> (.+))|(.+))} \
	       $k match type flags nlinks owner group size date i1 i2 i3 i4]
    #  set r [regexp {^([^ ])([^ ]+) (*[0-9]+) +([^ ]+) +([^ ]+) +([0-9]+) \
	#	  +(............) (((.+) -> (.+))|(.+))} 
    # $k match type flags nlinks owner group size date i1 i2 i3 i4]

    set nlinks [string trim $nlinks]

    if {$ftp($ftpI,debug) } {
      #	puts "back from try regular dir"
    }
    if {!$r} {
      #	puts "failed regular, trying andriod"
      frputs "Parse list " k
      set r [regexp {^([^ ])([^ ]+) ([^ ]+) +([^ ]+) (.*)\
			 ([0-9]{4}-[0-9]{2}-[0-9]{2}\
			      [0-9]{2}:[0-9]{2}) (((.+) -> (.+))|(.+))} \
		 $k match type flags owner group size date i1 i2 i3 i4]
      #                type  flags   owner group size date file link|file
      # if that worked, fix up the size...
      if {$r} {
	if {$type != "-" } {
	  set size 0
	} else {
	  set size [string trim $size]
	}
      }
    }
    frputs "Parse list result  " r date i1 i2 i3 i4 
    if {!$r} {
      #	puts "android failed..."
      # Try wuarchive.wustl.edu parsing
      if {$ftp($ftpI,debug) } {
	#	  puts "try wuarchive.wustl.edu dir"
      }
      set r [regexp {^([^ ])([^ ]+) (.*[0-9]+) +([^ ]+) +([0-9]+)\
			 +(............) (((.+) -> (.+))|(.+))} \
		 $k match type flags nlinks owner   size date i1 i2 i3 i4]
      if {!$r} {
	
	# Try WinNT parsing
	if { $ftp($ftpI,debug) } {
	  Log [_ "try WinNT dir"]
	}
	
	set \
	    r \
	    [regexp \
		 {(.................)(......................)(.+)} \
		 $k match date type i1]
	if {!$r} {
	  PopError [_ "Error parsing ftp LIST output: %s" $k]
	  return $dummy
	}
	set i3 {}
	set type [string trim $type]
	set flags rwxrwxrwx
	set nlinks 1
	set owner 0
	set group 0
	if {$type == "<DIR>"} {
	  set size 0
	  set type d
	} else {
	  set size $type
	  set type n
	}
      }
      set group 0
    }

    if {"$i3" != ""} {
      set file [string trimright $i3 "\n"]
      set link [string trimright $i4 "\n"]
    } else {
      set file [string trimright $i1 "\n"]
    }

    #	    if {"$file" == "." || "$file" == ".."} continue
    if {$type == "-"} { set type n}
    switch -exact $type {
      d  { set filetype fd }
      l  { 
	if { $config(ftp,fastlink) == 1 } {
	  set r [catch {FTP_IsDir $ftpI "$directory/$file"} outp]
	  frputs "FTP dir test  " r outp
	  set linkfound 1
	  if { $r != 0 } { PopError [_ "Fatal error: %s" $outp]; CleanUp 1 }
	  if {$outp == 0} {
	    set filetype fl
	  } else {
	    set filetype fld
	    set link $outp
	  }
	} else {
	  set filetype fld
	}
      }
      c  -
      s  -
      p  -
      n  { set filetype fn }
      default { PopError [_ "Error parsing ftp LIST output: %s" $k]; \
		    return $dummy }
    }
    set sec [FTPDateStringToSeconds $date]
    set tmp [switch -exact $config(fileshow,sort) {
      time      {format "%011d" $sec }
      rtime     {format "%011d" [expr 2147483647 - $sec]}
      size      {format "%011d" $size }
      extension {file extension $file }
      default   {expr { " "}}
    }]
    #     puts ">$tmp<"
    switch -exact $filetype {
      fn -
      fl  {set sortval 1$tmp}
      fd -
      fld {set sortval $sortval_d$tmp}
    }
    lappend dl [list $sortval $file $filetype $size \
		    $sec $flags \
		    $owner/$group \
		    $link $nlinks\
		    {} {}]
    lappend sortlist "$sortval $file"
  }
  if {$linkfound} {
    # make sure we leave the working dir in the right place
    catch {FTP_CD $ftpI $directory}
  }
  if {$ftp($ftpI,debug) } {
    #      puts "$dl"
  }
}
# Make sure link is open, don't open it if it is already open
proc OpenFTP { ftpI sftp} {
  global glob config env
  set ftpIleft ""
  set ftpIright ""
  set rl [regexp {([a-z]?)ftp://([^/]*)(.*)} $glob(left,pwd) \
	      match lsftp ftpIleft directory]
  set rr [regexp {([a-z]?)ftp://([^/]*)(.*)} $glob(right,pwd) \
	      match rsftp ftpIright directory]
  if {$ftpIleft == $ftpI || $ftpIright == $ftpI} {
    # Link already open
    return ""
  }
  Log [_ "Opening %sFTP connection to %s" $sftp $ftpI]
  set pwmess [expr {$sftp == "s" ?\
			[_ "\n(For SFTP multi part passwords\
                           ('password' 'private key file' 'Passphrase')\
                           \nplease enclose the whole entry\
                            in braces {} see 8.8 in the users guide)\n"] : ""}]
  # first see if we can find a match in the config(ftp,site_usage) rule list
  set r 0
  foreach k $config(ftp,login) {
    if {[string match [lindex $k 0] $ftpI]} {
      set user [lindex [lindex $k 1] 0]
      set passwd [decript $env(USER) [lrange [lindex $k 1 1] 1 end-1]]
      set proxy [lindex $k 2]
      set initcmd [lindex $k 3]
      if {$passwd == "XXX"} {
        set t [FTPEntryDialog \
		   [_ "%sFTP Login:0" $sftp] \
		   [_ "Connecting to %s: Please enter password%s" $ftpI $pwmess]\
		   $user]
        if {$t == ""} {
          error "ABORT_LOGIN"
        }
        set passwd [lindex $t 1]
      }
      if { $passwd == "" } {
        set passwd $config(ftp,password)
      }
      if { $proxy != "" } {
        set r [catch {FTP_OpenSession \
			  $ftpI \
			  $sftp \
			  $proxy \
			  $user@$ftpI \
			  $passwd\
			  $ftpI \
			  $initcmd} out]
        set glob(ftp,$ftpI,host) $proxy
        set glob(ftp,$ftpI,passwd) $passwd
        set glob(ftp,$ftpI,user) $user@$ftpI
      } else {
        set r [catch {FTP_OpenSession \
			  $ftpI \
			  $sftp \
			  $ftpI \
			  $user\
			  $passwd \
			  $ftpI \
			  $initcmd} out]
        set glob(ftp,$ftpI,host) $ftpI
        set glob(ftp,$ftpI,passwd) $passwd
        set glob(ftp,$ftpI,user) $user
      }
      frputs "back from first open " r out
      if { $r == 0 } {
	Log [_ "%sFTP connection to %s open" $sftp $ftpI]
	return
      } else {
	break 
      }
    }
  }
  while { 1 } {
    global errorInfo
    if {! [info exists user]} {
      set user [expr {$glob(os) != "WIN32" ? $env(USER) : $env(USERNAME)}]
    }
    set passwd $config(ftp,password)
    frputs "open error " r out
    if { $r && (! [string match -nocase "*password*" $out] && \
		! [string match -nocase "*connecting*" $out]) } {
      return -code error $out
    }
    if { !$config(ftp,anonymous) } {
      if {$r && $glob(debug) } {
	set mess "$errorInfo \n  [_ "Connecting to %s:\
                   Please enter username and password" $ftpI]"
      } else {
	if {! $r} {
	  set out {}
	}
	set mess [_ "%sConnecting to %s:\
                   Please enter username and password$pwmess" $out $ftpI]
      }

      set t [FTPEntryDialog [_ "%sFTP Login:1" $sftp] $mess $user]
      if {$t == ""} {
	error [_ "ABORT_LOGIN"]
      }
      set user [lindex $t 0]
      set passwd [lindex $t 1]
      if { $passwd == "" } {
	set passwd $config(ftp,password)
      }
    } else {
      if { $r } {
	set mess [_ "Error: %s Connecting to %s:\
                   Please enter username and password$pwmess" $out $ftpI]
	set t [FTPEntryDialog [_ "%sFTP Login:2" $sftp] $mess $user]
	if {$t == ""} {
	  error [_ "ABORT_LOGIN"]
	}
	lassign $t user passwd
      }
    }
    if { $config(ftp,proxy) != "" && $config(ftp,useproxy)} {
      set r [catch {FTP_OpenSession $ftpI \
			$sftp $config(ftp,proxy) \
			$user@$ftpI \
			$passwd $ftpI ""} out]
      set glob(ftp,$ftpI,host) $config(ftp,proxy)
      set glob(ftp,$ftpI,passwd) $passwd
      set glob(ftp,$ftpI,user) $user@$ftpI
    } else {
      set proxy {}
      set r [catch {FTP_OpenSession $ftpI \
			$sftp \
			$ftpI \
			$user \
			$passwd \
			$ftpI "" } out]
      set glob(ftp,$ftpI,host) $ftpI
      set glob(ftp,$ftpI,passwd) $passwd
      set glob(ftp,$ftpI,user) $user
    }
    if {$r} continue
    set ask [smart_dialog .apop . [_ "Login sucessful."] \
		 [list [_ "Login sucessful.\
		     \nDo you want to save the name and password for site: \
                  \n%s" $ftpI]] 0 2 [list [_ "No"] [_ "Yes"]]]
    if {$ask != 1} break
    set new {}
    foreach k $config(ftp,login) {
      if {![string match [lindex $k 0] $ftpI]} {
	lappend new $k
      }
    }
    set config(ftp,login) [lappend new [list $ftpI [list $user $passwd] $proxy]]
    FixPasswords
    
    break
  }
  
  Log [_ "%sFTP connection to %s open" $sftp $ftpI]
}
# Make sure link is closed, don't close if in use
proc CloseFTP { ftpI } {
  global glob config
  set ftpIleft ""
  set ftpIright ""
  set rl [regexp {[a-z]?ftp://([^/]*)(.*)} $glob(left,pwd) \
	      match ftpIleft directory]
  set rr [regexp {[a-z]?ftp://([^/]*)(.*)} $glob(right,pwd) \
	      match ftpIright directory]
  if {$ftpIleft == $ftpI || $ftpIright == $ftpI} {
    # Link in use
    return ""
  }
  #Log "Closing FTP connection to $ftpI"
  Try { FTP_CloseSession $ftpI } \
      [_ "Could not close FTP session nicely, (non-fatal)\n"] 1
  catch {unset glob(ftp,$ftpI,host)}
  catch {unset glob(ftp,$ftpI,user)}
  catch {unset glob(ftp,$ftpI,passwd)}
}
# Make sure link is closed, don't close if in use
proc CloseFTP { ftpI } {
  global glob config
  set ftpIleft ""
  set ftpIright ""
  set rl [regexp {[a-z]?ftp://([^/]*)(.*)} $glob(left,pwd) \
	      match ftpIleft directory]
  set rr [regexp {[a-z]?ftp://([^/]*)(.*)} $glob(right,pwd) \
	      match ftpIright directory]
  if {$ftpIleft == $ftpI || $ftpIright == $ftpI} {
    # Link in use
    return ""
  }
  #Log "Closing FTP connection to $ftpI"
  Try { FTP_CloseSession $ftpI } \
      [_ "Could not close FTP session nicely, (non-fatal)\n"] 1
  catch {unset glob(ftp,$ftpI,host)}
  catch {unset glob(ftp,$ftpI,user)}
  catch {unset glob(ftp,$ftpI,passwd)}
}
# This returns true if dir is an FTP dir and also sets ftpI and directory,
# in the callers context
# false if not and does not set the vars.

proc IsFTP { dir } {
  return \
      [uplevel regexp {\[a-z]?ftp://(\[^/]*)(.*)} \{$dir\} match ftpI directory]
}

