# -*- tcl -*-
set rev <20170326.1911.40>
# puts "loading frECF $rev"

# The "filerunner Exec Call Formater (frECF).
# the primary purpose of this function is to assemble an exec
# call that will do the "right" thing.
# It treats all parameters as strings, thus (hopefully) avoiding
# triping over special characters of meaning in the tcl syntax.
# But wait, the exec call needs to have special characters escaped
# the specials are "{} []$\". This appears to be true only for MS
# windows. To handle "native" file names we do our own quoting using
# FixFileName.

# frECF parameters:
#     fmat          A prototype of a format string (see below)
#     parms         A list of strings (file names) to convert (format)
#                   This should ALWAYS be a list. If one file, pass
#                   [list <file>]. This protects us (you) from
#                   odd ball strings in <file>
#     args          Options detailing various things. (see below)
#
#  effectively the frECF function calls 'format' with the given string
#  and set of parms and {optionally) executes the result.
#
# Details...
# In general 'format' requires one '%' for each parm.  The proto type
# string 'fmat' may be deficient in this regard due to more entries
# in the 'parms' list than '%'s in 'fmat'. This is corrected by
# adding '%' factors to the string. In the resulting processed
# 'fmat' string (pfmat) we recognize:

#  %s    parm converted with 'file native'
#        This does exactly nothing on other than MS windows systems
#  %m    converts to a %s with its parm converted with a function
#        provided by the 'args' -m option.
#  %n    Converts to a simple %s i.e. no special conversions
#        Read it a "normal"
#  %a    Replace this with default ( "%[ns]"). This is used if exactly
#        one conversion is needed at this point in the string.
#  %b    Same as %a, except put as many as needed here to get the
#        required number of conversions. (usually followed by & or |)
#  %q    same as %s except enclose in \" (instead of ").
#  %r    same as %s but no enclosing quote at all
#  %%    converts to %
#  %"    converts to "
#  %     (percent space) converts to "" nil, nothing.

#  Positional conversions are allowed (i.e. %2@s %1@s etc.) except for %a & %b.
#  We use "@" instead of "$" to avoid tcl's desire to dereference the "$".
#  Special handling of quotes is done as follows:
#  A 'chunk' of fmat is defined as a bit that is delimined by 'white space'
#  and (possibly) the end of the string. For each chunk that ends up
#  containing one or more conversions (i.e. % signs) we insure that the
#  resulting chunk is in quotes.  If they are already there we do nothing.
#  If not they are added. This allows the conversions to be concationated
#  with other string fragments as desired. (This is needed, for example,
#  to build a URL for with a file name, i.e. "file://%s".) 'chunks' that
#  contain no conversions are passed without modification as is all
#  white space. This is done to
#  allow constructs like "( file file) " which also requires redefining
#  the default conversion: {"" ( %b ) ""} [list -d { \"%\"}]
#  We use "stringGetToken" from the "slLib.tcl" to find and get
#  'chunk's.
#
#  args is a list consisting of -<option type>  <option value>
#       args is processed such that you can pass one or more lists and
#       or a series of individual parameters. [concat {*}$args]
#
#  defined args are:
#  -m       <function>    This function will be called to convert parms
#                         for '%m' pfmat specifiers.

#  -default <conversion>  <conversion> will be added to the 'pfmat'
#                         at '%a' and/or at the end of the pfmat as
#                         many times as need to get the required
#                         number of conversions. If '-d' is not given,
#                         { "%n"}} is used for MS windows, else { "%s"}.
#                         If a '%b' is encountered it is assumed that
#                         there are no more '%' and it will be replaced
#                         with as many of these conversions as needed
#                         to complete the format string '%' count.
#                         The first %b conversion will complete the
#                         'chunk' it is found in, including any trailing
#                         characters after the %b in its 'chunk'.
#                         Then, if needed, additional 'chunk' is started
#                         with the characters preceeding the % in
#                         the 'default'. Each of these 'chunk's will
#                         be completed with the characters following
#                         the conversion character in 'default'. This
#                         string will ususlly need to be protected with
#                         the list command, e.g. [list { "%n"}]

#  -fonly         1       the result of the format call is returned
#                         without being executed.

#  -background    1       execute asynchronously (done by 'after' command).

#  -alterr  <message>     <message> is an alternative error message
#                         to be used on execution failure. <message>
#                         may include up to 3 '%s's which will be replaced 
#                         with as follows:
#                         1 %s the error returned from the command
#                         2 %s the command and the error returned
#                         3 %s the command, ::errorInfo and the output
#                         The default <message> is:
#                         "While executing: \n%s\n%s".
#                         If <message> is {}, no error are reported.
#                         Errors are passed to 'PopError' except if:

#  -error   <efun>        Errors are passed to the function '<efun>'
#                         If <efun> is {} no errors are reported as such,
#                         but the -post function (if one) is called.   

#  -post    <pfun>        function to call with result of the command
#                         this is most useful for async commands. 
#                         <pfun> will only be be called if the command
#                         produces no errors. The variable name 
#                         containing the output (if any) will be  
#                         appended to <pfun> for this call (call by name).
#

# The pfmat string is sequentially processed replacing % specifiers
# with converted files from the parms list until it is exhausted.
# Any remaining pfmat parts are processed to remove left over %
# specifiers.
#
# There are two types of errors:
# 1) those that come from a bad call and
# 2) those that are caused by attempts to execute and external
#    function which fails
# Type 1) errors are reported with the tcl 'error'
# Type 2) are reported with the -error function
#
# frECF returns:
#  A 2 or 3 element list: {<r> <output> ?<command>?}
#  if <r> is zero,     then there was no error and <output> is the result
#  if <r> is non-zero, then there was an error and <output> is the
#                      error message and <command> is the command that
#                      generated the message

proc frECF {pfmat {parms {}} args} {
  frputs "frECF: entry  " pfmat parms args
  # set up defaults
  set alterr "While executing:\n%s\n%s"
  set error "PopError"
  set fonly 0
  set default $::frECF::default
  set m {format %s }
  set n {file nativename }
  set background 0
  set post {}
  set nop {}
  set possible {-alterr -error -fonly -default\
		    -m -n -background -post -nop}
  
  foreach {opt value} [concat {*}$args] {
    set tar [lsearch -glob -all -inline $possible "$opt*"]
    if {$tar == -1 || [llength $tar] != 1} {
      error "frECF: option ($opt) must be one of $possible"
    }
    set [string range $tar 1 end] $value
  }
  if {[string match *%a* $default] ||\
	  [set ind [string first "%" $default]] == -1} {
    error "frECF: -default ($default) must\
           have NO \"%a\" and at least 1 \"%\""
  }
  set len [llength $parms]
  frputs "frECF:  in: " pfmat
  # Here we define the prefix conversion
  # This should be some separator character(s) and a quote 
  set pre [string range $default 0 $ind-1]
  set pos [string range $default $ind+2 end]
  
  frputs "frECF: pre:  " pre pos pfmat parms len
  # I explored two ways to go from here:
  # 1) Change 'pfmat' to be a proper 'format' string and call 'format'
  #    which means:
  #    a) Changing all the %a to the current default and
  #    b) Processing the %b to put the needed default conversions
  #       either at the %b spot or at the end of the string.
  #    c) Preprocess each occurance of %n or %s or %m to convert
  #       the 'parms' list into a new, converted list while changing
  #       each entry in the 'pfmat' string to %s.
  #    d) Calling 'format' with the string and the processed 'parms'
  #
  #  or
  #
  #  2) Process the 'pfmat' string and 'parms' in one pass:
  #     a) Find each '%' and do the required conversion of
  #        'parms' directly into the resulting string.
  #
  #  While 2) sounds easier, it really isn't, at least in the
  #  amount of code needed.  However, it allows us to be more
  #  forgiving around the conversion of positional specifiers.
  #  While 'format' requires all or none in this reguard we only
  #  require not (explicitly or implicity) asking for the same
  #  element twice. (As each conversion is processed we mark it
  #  as done. If no particular element is requested we
  #  take the first un-done element.  If subsequently that
  #  element is requested, we throw an error.)
  

  # now build the command...
  set start 0
  set used {}
  set this {}
  set foundB 0
  set begin 0
  # Here is a trick to get the parms out in case there are no "%"s
  if {[string index $pfmat end] == "&"} {
    set pfmat [string replace $pfmat end end " %b &"]
  } else {
    append pfmat " %b"
  }
  while {[set endc [stringGetToken $pfmat start]] >= $start} {
    # This may be empty, but it preserves white space
    append cmd [string range $pfmat $begin $start-1]
       # if {[incr w1] > 10} {puts "w1 break"; break}
    # if token ends with '%' the next char should be a space
    set chunk [string range $pfmat $start $endc]
    if {[string index $pfmat $endc] == "%"} {
      set chunk [string replace $chunk end end]
      incr endc
    }
    set begin [incr endc]
    frputs "frECF  " chunk pfmat cmd
    # do the %" here, does not depend on whats on
    # either side. Can not do %% this way as %%s would then
    # look like %s.
    set chunk [regsub -all {%\\?\"} $chunk \" ]
    set startC 0
    set matchQ 0
    set perCfound 0
    while {[set loc [string first % $chunk $startC]] != -1} {
      # if {[incr w2] > 10} {puts "w2 break"; break}
      if {[string index $chunk $loc-1] == "\\"} {
	set startC [incr loc]
	continue
      }
      set this [string index $chunk $loc+1]
      # flush everything up to the '%'
      append cmd [string range $chunk $startC $loc-1]
      # double % flush it here and restart the chunk with out
      set indx -1
      if {[string match \[0-9\] $this] && \
	      [string index $chunk $loc+2] == "@"} {
	# this is the index + 1
	set indx [incr this -1]
	set this [string index $chunk $loc+3]
	incr loc 2
      }
      if {$this == "%" || $this == { } || [llength $used] >= $len } {
	# start set chunk to what remains
	set chunk [string range $chunk $loc+2 end]
	if {$this == "%"} {
	  append cmd "%"
	}
	set startC 0
	continue
      }    
      # Must be an actionable %
      incr perCfound
      set closeq {}
      if {$this in {a b}} {
	append cmd $pre
	set closeq $pos
      } elseif {$this == "q"} {
	set closeq {\"}
	append cmd $closeq
      } elseif {$this in {s m n}} {
	set closeq "\""
	append cmd $closeq
      }
      set startC [expr {$loc + 2}]
      # At this point 'this' is the conversion char.
      # 'startC' is the next char. in this 'chunk' and
      # preamble chars from pfmat have been moved to 'cmd'
      set firstB 0
      if {$this == "a" || $this =="b" || $foundB} {
	if {$this == "b"} {
	  set foundB 1
	  set firstB 1
	}
	set this [string index $default $ind]
      }
      frputs "frECF  "  indx this startC loc foundB used cmd pre
      while {1} {
      # if {[incr w3] > 10} {puts "w3 break"; break}
	if {$indx == -1} {
	  while {[incr indx] in $used} {}
	} elseif {$indx in $used} {
	  error "frECF: Conversion # [incr indx] used twice."
	} elseif {$indx >= $len || $indx < 0} {
	  error "frECF: Conversion # [incr indx] has no value"
	}
	lappend used $indx
	switch $this {
	  q -
	  r -
	  s {append cmd [FixFileName [file native [lindex $parms $indx]]]}
	  m {
	    set r [catch {uplevel "{*}$m [lindex $parms $indx]"} out]
	    if {$r == 0} {
	      append cmd $out
	    } else {
	      # this might be a type 2 error... but
	      error "frECF: \"m\" conversion error $out"
	    }
	  }
	  default {append cmd [FixFileName [lindex $parms $indx]]}
	}
	append cmd $closeq
	# Spilling out the remaining parms?
	if {!$foundB || [llength $used] >= $len} {
	  # nop, either not doing or done, break the while {1}
	  if {$foundB && !$firstB} {
	    # This is the final closing of a %b series
	  }
	  break
	}
	set indx -1
	append cmd $pre
      } ; # End of while(1)
      # We need to close this conversion...
    }   ; # End of while on chunck
    
    if {$perCfound != 0} {
      # if we found a % append the final bit
      append cmd [string range $chunk $startC end]
    } else {
      # Otherwise, pass the whole chunk
      append cmd $chunk
    }
    set start $begin
  }
 # End of main While...getToken
  
  append cmd [string range $pfmat $begin $start-1]
  # If a conversion was already quoted we may have double quotes.
  # Here quick releaf
  set cmd [regsub -all {([^\\]")"} $cmd {\1}]
  # set cmd [format $new {*}$pparms]
  frputs "frECF: " pfmat
  frputs  parms 
  # frputs pparms 
  frputs cmd
  frputs post
  frputs background
  
  if {[string is true -strict $fonly]} {
    return $cmd
  }

  if {$background} {
    after 10 "frECF_Async [list $cmd] \
                        [list $alterr] \
                        [list $error] \
                        [list $post]"
    return 0
  } else {
    return [frECF_Async $cmd $alterr $error $post]
  }
}
proc frECF_Async {cmd alterr error post} {
  set r [catch {uplevel $cmd} out]
  frputs "Async frECF:  " cmd out r
  if {$r != 0 && $error != {} && $alterr != {}} {
    switch [stringCount "%s" $alterr] {
      3 {set out [format $alterr $cmd $::errorInfo $out]}
      2 {set out [format $alterr $cmd $out]}
      1 {set out [format $alterr $out]}
    }
     if {[catch {{*}$error "$out"} outt] != 0} {
      error "frECF: Calling $error: $outt"
    }
  } else {
    if {$post != {}} {
      catch {{*}$post out} hun
      frputs "frECF After post call  " hun ::errorInfo
    }
  }
  return [list $r $out $cmd]

}


# To keep the run time simple we set up some stuff at load time
namespace eval frECF {
  variable default { "%n"}
}

if {$::tcl_platform(platform) == "windows"} {
  set ::frECF::default { "%s"}
}
