# <20160831.1440.22>
# This bit of tcl is an attempt to put together a way to run child programs
# while still haveing a LOT of control over them.
#
# We use the tcl 'open |..' command to open a pipe to the given program.
# depending on the options given we then:
#   allow writing to the pipe to pass info to the program
#   provide a call back capability to process info from the program
#   provide a kill call to eliminate the program if needed.
#
# Commands provided:
#
# pipeoExec command options ?call back functions?
#
#           command is the command passed to open with a "|" in front
#                   it opens a pipe with stdin and stdout being here
#                   This command must direct stderr as desired.
#                   Stdout will be readable (option r). Possible options:
#                   2>/dev/null 2>@1 2>&1 (see tcl exec man page).
#                   (see below how to access these)
#           options is the option(s) passed to open. 'r' for read only
#                   "r+". Other options are possible, see open(n).
#           call back functions is a list of one or two functions
#                    which are called with info on the progress of
#                    the pipe. If only one function is given it will
#                    be called with all info.  It two are given the
#                    first (log) will be called with log information and
#                    the second (data) with data from the pipe.
#
#       pipeoExec returns a list of two values:
#                 if the first is 0, the second is the pipe 'handle'
#                 if the first in !0, the second is the error information
#                                     on why the pipe was not opened.
#
#       The data output (send to the data call back) will be one of:
#                'handle' data <data from the pipe>   ; data packet
#                'handle' eof                         ; eof encounterd
#                                                       or pipe is aborted, 
#                                                       pipe is closed
#       The log call back will be one of:
#                'handle' abort      ; if the pipe aborts for what ever reason
#                'handle' done       ; pipe completes, it is closed, 
#                                      eof message sent 
#                                      to the data call back
#                'handle' kill       ; kill was called on the pipe
#                'handle' kill-soft  ; soft-kill was called on the pipe
#
# puts handle <data to write>        ; handle is what was passed back on 
#                                     pipeoExec call
#
# pipeoSoftAbort handle   ; or      ; Sends SIGINT to the pipe process(es)
# pipeoAbort handle 1  
#
# pipeoAbort handle 2 (default is 2); Aborts the pipe process(es) 
#                                    (kill -15, kill -9)
#                                   ; and closes the pipe, 
#                                     will force the call backs

# Usually you may have to supply proper termination on lines sent to the pipe
# (\n or \r\n) or. But see the fcontrol(n) man page. 'handle' is the
#  channelId if you want to call fcontrol to set this.

# We use Tclx for kill.  We want an equivalent for windows so we won't try
# to write our own... well, actually we did. See killCode.tcl

namespace eval pipeoExec {
  variable pipeoExParms {}
}


proc pipeoExec { cmd options {callback {puts puts}} } {
    
  if {$::tcl_platform(platform) == "unix"} {
    set cmd [regsub -all {\\} $cmd {\\\\}]
  } else {
    set cmd [regsub -all {\\} $cmd {\\}]
  }
  set r [catch {open "| $cmd" $options} fid]
  frputs "pipeoExec open  " cmd r fid 
  if {$r == 0} {
    # If only one callback function, double it so we don't need to
    # check later
    if {[llength $callback] == 1} {
      lappend callback $callback {}
    }
    lappend ::pipeoExec::pipeoExParms [list $fid $callback]
    fconfigure $fid -buffering line
    fconfigure $fid -blocking 0
    fconfigure $fid -translation auto
    # schedule the completer...
    set rr [catch {chan event $fid readable "pipeoComplete $fid"} out]
    frputs "pipeoExec: open success completion set  " rr out
  }
  return [list $r $fid]
}

proc pipeoComplete {fid} {
  set id [lsearch -exact -index 0 $::pipeoExec::pipeoExParms $fid]
  if {$id == -1} {
    error "pipeoComplete called but $fid is not an active channel"
  }
  # A bit of a problem we address here.  If the other side is fast
  # by the time we get around to reading, there may be more than
  # one line waiting.  Not only that, it is possible that the
  # last part of such a line is a fragment of the next line.
  # So, we split the line into records ending in "\n" here
  # and put any remains as the third entry in the pipeoExParms list
  
  lassign [lindex $::pipeoExec::pipeoExParms $id 1] log rtnv remains
  # eval  {*}$log $fid entered
  #Log [_ "%s called" CompleteShell_pipeo]
  set out ""
  set r [catch {set out [read $fid]} ]
  #frputs "pipeoComplete entered  " fid out
  # add the remains...
  set out $remains$out
  if {$out != ""} {
    # Here we split the lines on "\n".  Must use string code to
    # avoid nasty embeded special chars like {} [] and $ .
    set fc 0
    while {[set eos [string first "\n" $out $fc]] != -1} {
      
      eval {*}$rtnv $fid data {[string range $out $fc $eos]}
      set fc [incr eos]
    }
    lset ::pipeoExec::pipeoExParms [list $id 1 2] [string range $out $fc end]
	# [lreplace $::pipeoExec::pipeoExParms $id $id \
	#      [list $fid [list $log $rtnv [string range $out $fc end]]]]
    if {$remains != {} || [string length $out] != $fc} {
      frputs "PipeoComplete:  " remains "extra: =[string range $out $fc end]  " \
		  ::pipeoExec::pipeoExParms
    }
  }
  if {$r || [eof $fid]} {
    # in case we have left over chars in remains...
    if {$remains != {}} {
      eval {*}$rtnv $fid data {$remains}
    }
    # do we need this???
    # if {[$w.text get "end - 1 chars"] == "\n"} {
    #  $w.text delete "end - 1 chars"
    # }
    # Maybe this is better...
    set ::pipeoExec::pipeoExParms [lreplace $::pipeoExec::pipeoExParms $id $id]
    catch {close $fid}
    eval {*}$rtnv $fid eof
    if { $r } {
      eval {*}$log $fid abort
    } else {
      eval {*}$log $fid done
    }
  }
}

proc pipeoSoftAbort {fid} {
  pipeoAbort $fid 1
}

proc pipeoAbort {fid {soft 2}} {
  set id [lsearch -exact -index 0 $::pipeoExec::pipeoExParms $fid]
  if {$id == -1} {
    frputs "pipeoAbort called but $fid is not an active channel "
    return
  }
  lassign [lindex $::pipeoExec::pipeoExParms $id 1] log rtnv
  set s [expr {$soft == 0 ? "" : "-soft"}]
  eval $log $fid kill$s
  set pid [pid $fid]
  # We do both the sigterm and the kill, just to make sure...
  # if there is no tcl kill the system program will be called
  # it has a different parameter set to id the signal...
  kill $pid
  catch {close $fid}
  pipeoComplete $fid
}

