# <20170115.1811.11>

# The rsync command makes every attempt to make the copy using direct
# requests to the target or source machine.  If both files are local
# things are real simple.
#
# We translate local file names to host:file names for the following
# connections:
#
# nsf
# cifs
# ftp
# sftp
#
# nsf works well out of the box. For the others we need to translate
# the root address from what it looks like here to what it is on the
# other machine. We also can only handle transfers where at least one 
# end is local.
#
#
#
# This procedure takes a file name and returns the components
# needed to build a rsync reference to an external file.
#
# Possible file name references:
#
# <file>                           (local file)
# [user@]host:file                 (rsync using a remote shell & scp's format)
# [user@]host::<file>              (rsync using an rsync daemon)
# rsync://[user@]host[:port]<file> (rsync using an rsync daemon)
# 
# If the file is local, a list containg an empty (i.e. {}) value
# followed by the passed in name.
#
# If the reference is to a remote file that is mounted on this
# system the return consists of a list containing the remote 
# system name (host) followed by the file name as it appears on the remote
# system.
# 
# The intent here is that the caller can use the file name thusly:
#
# rsync <options> [join [fileNameTranslate <souce file>] :]
# or, for more complex file names:
# lassign [fileNameTranslate <souce file>] host file
# if {$host == {}} {
#   set url $file
# } else {
#   set url [format "rsync://%s@%s:%s/%s" $user $host $port $file]
#
proc fileNameGetParm {key string {noneIsM1 0}} {
  set st [split $string ,]
  set ind [lsearch -regexp $st "^${key}($|=.*)"]
  if {$ind < 0} {return [expr {$noneIsM1 ? -1 : {}}]}
  lassign [split [lindex $st $ind] =] nkey val
  return $val
}


# fileNameTranslage returns a list of doublets in a format suitable for 
# 'array set'..  The following will only be present if needed:
#
# scheme {file nfs cifs}
# path   for file -$sfile, for nfs target file, for cifs part of $sfile to add
#                                                        to share location
# host   for file not needed, for nfs|cifs host ip 
# user   for file not needed, for nfs|cifs user used in fstab
#
# share  only for nfs & cifs the share name (host:/??) or (//host/???)
# If a vaue is unknown it will be {}.
#
proc fileNameTranslate {sfile} {
  # We do not require the file to actually exist but we do assume
  # that the path is real and on this system
  set scheme {}
  regexp -- {^([a-z0-9+.-][a-z0-9+.-]*):} $sfile dummy scheme
  #puts "$sfile $scheme"
  if {$scheme != {}} {
    # this is a url of some sort.  Not sure what to do with it...
  } else {
    set mtab "/etc/mtab"
    if {![info exists ::fileNameTranslate::mtime]} {
      namespace eval ::fileNameTranslate {
	variable maps {}
      }
    }
    set cntrl [split [read [open $mtab r]] \n]
    set ::fileNameTranslate::maps {}
    # external file systems are either <name>:* or <ip>:* or //<name>/*
    # build a list of possible mount points for external file systems
    foreach ln $cntrl {
      lassign $ln sys mp dm ops
      if {$sys == {} || [string range $sys 0 0] == "#"} {continue}
      if {[string range $sys 0 1] == "//" || [string match "*:*" $sys]} {
	#puts "$sys $mp $dm $ops"
	lappend ::fileNameTranslate::maps \
	    [list $mp $sys host [fileNameGetParm addr $ops] \
		 user [fileNameGetParm username $ops]]
      }
    }
    set ::fileNameTranslate::maps \
	[lsort -decreasing -index 0 $::fileNameTranslate::maps]

    # Here we have the control base, it is time to see if the passed in
    # file is remote
    foreach rmt $::fileNameTranslate::maps {
      # remains will be host and user
      set remains [lassign $rmt key ho]
      set keyLn [string length $key]
      if {[string compare -length $keyLn $key $sfile] == 0} {
	# this would be it, the longest match since the map is desending
	# The nfs file system is an easy translation.  The cifs, not so much.
	if {[string range $ho 0 1] != "//"} {
	  # nfs...
	  return \
	      [list path \
		   [regsub {^[^:]+:(.*)} \
			$ho {\1}][string range $sfile $keyLn+1 end] share $ho\
		   {*}$remains scheme nfs]
	} else {
	  return [list path [string range $sfile $keyLn+1 end] share $ho\
		      {*}$remains scheme cifs]
	}
      }
    }
    # If we get here its a local file
    return [list path $sfile scheme file]
  }
}

# frTranslateDir does the following:
# 1) breaks up the file/dir name into the local and remote parts
#    For remote names it scans the rsync look up table at config(rsync)
#    Each entry in this table should provide:
#    0) the share name as in /etc/fstab (first entry on the line)
#       if the share is ftp or sftp, then (s)ftp://user@host
#       (user is required as different users may have different dirs.)
#    1) for other than nfs, the dir this connection accesses on the host
#       for example, for ftp this might be /home/ftp/pub 
#    2) rsync mode to use, one of {shell daemon}
#    3) rsync options to use with this machine (if not provided the 
#       global rsync options are used.
#    4) Here is the run down on passwords:
#       nfs, remote shell or daemon if needed (user name from fstab)
#       cifs, remote shell or daemon if needed (user name from fstab)
#       ftp,  remote shell or daemon only if different from ftp connection
#       sftp, remote shell or daemon only if different from sftp connection
#
#       We assume the remote shell option will use ssh and provide code
#       to pass the password.

# 
proc frTranslateDir {dir} {
  global glob config
  # if the 'dir' is ftp or sftp we can translate. At this time we don't 
  # do ?ftp for ? != "" or "s".
  #
  if {[IsFTP $dir]} {
    regexp {([a-z]?)ftp:.*} $dir match type 
    if {$type != {} && $type != "s"} {
      error "rsync command can not handle ${type}ftp files."
    }
    # it is a legal s?ftp path.  IsFtp set 'ftpI' and 'directory' for us
    lassign [split $glob(ftp,$ftpI,host) :] host port
    set parts [list scheme ${type}ftp \
		   path $directory\
		   user $glob(ftp,$ftpI,user)\
		   host $host \
		   port $port \
		   pwd $glob(ftp,$ftpI,passwd)]
  } else {
    set parts [fileNameTranslate $dir]
    lappend parts pwd {}
  }
  # at this point we have all we need for nfs and local files
  # for others we need to use a table (hopefully) set up in the
  # config area to translate the share or (s)ftp path to the root on 
  # the remote machine.
  frputs "parts of the target  " parts
  array set urlA $parts

  # Set the default options and password values
  upvar rsyncOps  rsyncOps
  upvar rsyncPwd  rsyncPwd
  upvar rsyncHost rsyncHost
  upvar rsyncMode rsyncMode
  set searchOps [list -index 0 -inline]
  # this while is just so the switch default can 'break'
  # it should NEVER loop.
  while {1} {
    # If local file, don't set the Pwd Host and Ops
    if {$urlA(scheme) == "file"} {return $urlA(path)}
    # Time to set option defaults
    # set rsyncHost {}
    #lassign [lindex $config(rsync,table) 0] d d rsyncMode rsyncOps rsyncPwd
    switch -exact $urlA(scheme) {
      ftp -
      sftp {
	set qPort [expr {$urlA(port) != {} ? ":$urlA(port)" : {}}]
	set pattern $urlA(scheme)://$urlA(host)$qPort
      }
      nfs -
      cifs {set pattern $urlA(share);lappend searchOps -nocase}
      default {error "rsync unknown file type: $urlA(schemd)"}
    }
    # At this point we know the file is remote
    set info [lsearch {*}$searchOps $config(rsync,table) $pattern]

    # don't require a table entry for 'nfs' files.
    if {$info == {}} {
      if {$urlA(scheme) == "nfs"} {break}
      error "rsync info for remote $pattern not in rsync table."
    }
    set rsyncPwd [decript $::env(USER) [lrange $rsyncPwd 1 end-1]]
    lassign $info key rPath Mode Ops Pwd
    set Pwd      [decript $::env(USER) [lrange $Pwd 1 end-1]]
    set rsyncMode [expr {$Mode == {} ? $rsyncMode : $Mode}]
    set rsyncOps [expr {$Ops == {} ? $rsyncOps : $Ops}]
    frputs "pwd  " Pwd urlA(pwd) rsyncPwd
    set rsyncPwd [expr {$Pwd != {} ? $Pwd : \
			    [expr {$urlA(pwd) != {} ? $urlA(pwd) : $rsyncPwd}]}]
    #set rsyncPwd [decript $::env(USER) [lrange $rsyncPwd 1 end-1]]
    frputs "decript  " rsyncPwd urlA(path)
    if {$urlA(scheme) != {nfs}} {
      set urlA(path) [lindex $info 1]/$urlA(path)
    }
    break
  }
  # Modules (see rsync man page) are indicated by mode "daemon=modname"
  lassign [split $rsyncMode =] dae module
  # All that is left is to build the rsync dir string. Note we do not 
  # put the port here. If needed it should be part of rsyncOps
  set rsyncHost $urlA(host)
  frputs "build name  " urlA(user) urlA(host) urlA(path) rsyncHost rsyncPwd
  return "$urlA(user)[expr {$urlA(user) != {} ? "@" : ""}]$urlA(host)[expr \
                           {$dae == {daemon} ? {::$module} : {:}}]$urlA(path)"
}

proc myLog {w mess} {
  # all these messages go to the window...
  # this code tries to make the window act as a terminal would
  # this means we need to interpret \r so as to replace lines
  # to do this we need to have some notion of the fact that a line
  # ended with a "\r". So as not to require global storage we
  # do this by moving 'insert' to the required location for
  # the delete, but we don't do that until we get the next line.

  set mess [regsub -all \r\n $mess "\n"]
  set mes [split $mess "\r"]
  #set mes lassign $mes ms
  set first 0
  lassign [$w.text yview] start end
  # puts "$start $end"
  foreach ms $mes {
    incr first
    lassign [split [$w.text index insert] .] insertL insertC
    if {$ms == {} || $first != 1} {
      $w.text mark set insert $insertL.0
      if {$ms == {}} {continue}
    }
    # do we have a pending delete?
    if {$insertC == 0} {
      $w.text delete $insertL.0 end-1c
    }
    $w.text insert insert $ms
  }
  if {$end == 1.0} {
    $w.text see insert
  }
}


proc CmdRsync {} {
  global glob config
  lassign [lindex $config(rsync,table) 0] d d rsyncMode rsyncOps rsyncPwd
  set rsyncHost {}
  foreach pane {left right} {
    set sel [$glob(listbox,$pane).file curselection]
    if {$sel != {}} {
      # first translate the from and to directories.
      foreach way {from to} pan "$pane [Opposite $pane]" {
	set ${way}Dir [frTranslateDir $glob($pan,pwd)]
	set ${way}Remote [string match $glob($pan,pwd) {*:*}]
      }
      if { $glob(left,pwd) == $glob(right,pwd) } {
	SandDsame 
	return {}
      }
      foreach file $sel {
	lappend fileList $fromDir/[lindex $glob($pane,filelist) $file 1]
      }
      lappend fileList $toDir
      
      # An alternate program can be indicated by rsyncMode "use=alt"
      lassign [split $rsyncMode =] op alt
      set prog [expr {$op == {use} ? "$alt" : "rsync -s"}]
      frputs "prog  " op alt prog
      set nill {}
      set w [ViewString [_ "%s progress window" $prog] nill ""] 
      # puts "window $w"
      # using expect package to help with password...
      catch {unset spawn_id}
      if {![info exists glob(expect)]} {
	# not defined => not loaded, 1 => full expect, 0 => Tcl expect
	set glob(expect) 1
	if { [catch "package require Expect" out] != 0 || $out == "0.0.0"} {
	  Log "Full Expect not found, using pure Tcl expect"
	  PopInfo [_ "Expect not found. Rsync will likely fail to properly\
                        login.\nPlease down load and install the 'expect'\
                        package from your\nsystem repo."]
	  set glob(expect) 0
	}
      }
      exp_log_user $glob(debug)
      if {$glob(debug)} {
	if {$glob(expect) == 0} {
	  # only my expect can handle a Log procedure.
	  exp_log_file -a -proc Log
	} else {
	  exp_log_file  -a -noappend $glob(tmpdir)/Log
	}
      }
      # ready to rock...
      # set cmd [subst -noc -nov [frECF [list spawn -noecho {*}$prog {*}$rsyncOps ]\
      # 				    $fileList\
      # 				    [list -f 1]]]
      set cmd [frECF [list spawn -noecho {*}$prog {*}$rsyncOps ]\
		   $fileList\
		   [list -f 1]]
       # set cmd [ReSpaceString "spawn -noecho $prog $rsyncOps" $fileList]
      frputs "spawn  " cmd
      set r [catch {{*}$cmd} out]
      if {$r != 0} {
	destroy $w
	return -code error "Really bad error: $out" 
      }
      set geo [buildStop $w [_ "Rsync: Running \
                     \n%s %s %s" $prog $rsyncOps $fileList]\
		   "expectAbort $spawn_id" 1]
      if {[winfo exists $w]} {wm geo $w $geo}

      # at this point we turn the rest over to a handler...
      frputs "sending:  " w spawn_id rsyncPwd rsyncHost
     
      after idle "rsyncExpectHandler $w $spawn_id \"$rsyncPwd\" $rsyncHost"
      return
    }
  }
}

proc expectAbort {spawn_id args} {
  # get here if user clicks the stop button
  if {[catch {exp_pid -i $spawn_id} pid] != 0} {
    frputs "expectAbort  " pid
    return
  }
  set sig [expr {$args == 1 ? "SIGINT" : "SIGKILL"}]
  catch {exec [auto_execok kill] -s $sig $pid} s
  frputs "kill  " sig pid s
}

proc rsyncExpectHandler {w spawn_id rsyncPwd {rsyncHost {}}} {
  frputs "received:  " w spawn_id rsyncPwd rsyncHost
  global glob
  lassign $rsyncPwd Pword none passphrase
  expect_after timeout \
      {set expect_out(1,string) "Connection timed out $rsyncHost";\
	   set re 1}\
      eof {set expect_out(1,string) "Connection closed $rsyncHost" ;\
	       set re 0}
  expect -re "(.*assword:.*)" \
      {myLog $w "$expect_out(1,string)"
	Log "sending password"
	frputs "open password  " rsyncPwd
	myLog $w "sending password\n"
	exp_send $rsyncPwd\r
	exp_continue} \
      -re "(.*.?assphrase for key .*: )" {
	myLog $w "$expect_out(1,string)"
	myLog $w "sending passphrase\n"
	exp_send "$passphrase\r"
	exp_continue} \
      -re "(.* host key is not .*y/n. |.* authenticity of host .*\(yes/no\)\? )"\
      {
	#set st [regsub -all {\r} $expect_out(1,string) {}] 
	myLog $w $expect_out(1,string) 
	incr ignorto
	if { [smart_dialog .apop . [_ "Accept new host?"] \
		  [list {} "$st" [_ "\nClick your answer."]] \
		  1 2 [list [_ "No"] [_ "Yes"]]] == 1} {
	  set an [expr {[string match "*(yes/*" $expect_out(1,string)] ? \
			    "yes" : "y" } ]
	  unset ignorto
	  frputs "back from y/n " an
	  myLog $w "sending $an\n" 
	  exp_send "$an\r"
	  exp_continue 
	} else {
	  myLog $w "sending 'no'\n" 
	  exp_send "no\r"
	  Log "Aborting Login"
	  set re 10
	}
      } \
      -re "(.*yes\r\n)" {
	myLog $w $expect_out(1,string)
	exp_continue} \
      -re "(.*Warning: .*)\r?\n" {
	myLog $w "$expect_out(1,string)"
	exp_continue}\
      -re "(.*Connecting .*)" {
	myLog $w "$expect_out(1,string)"
	myLog $w "sending password2\n"
	exp_send $rsyncPwd\r
	exp_continue} \
      -re "(.*Permanently added .*)\r*\n" {
	myLog $w "$expect_out(0,string)"
	exp_continue}\
      -re "(.*assword.*|.*unable.*|.*assphrase.*|.*onnection closed.*)" {
	myLog $w "exit open: Failed with $expect_out(1,string)\n"
	set re 5}\
      -re "(.*)\r*\n" {
	#frputs expect_out(1,string) expect_out(0,string)
	myLog $w "$expect_out(0,string)"
	exp_log_user $glob(debug)
	exp_continue}\
      -re "(.*)\r" {
	#frputs expect_out(1,string) expect_out(0,string)
	myLog $w "$expect_out(0,string)"
	exp_log_user $glob(debug)
	exp_continue}
  # We get here on:
  # close            (re = 0)
  # timeout          (re = 1)
  # password failure (re = 5)
  # reject new host  (re = 10)
  #
  # After a close/timeout we need to do a wait to get the 
  # status of the completion...m
  StopButRemove $w
  catch "exp_close -i $spawn_id"
  #ForceUpdate $inst
  set rtn [exp_wait -i $spawn_id]
  lassign $rtn r1 r2 r3 r4
  if {$r3 == 0 && $r4 == 0} {return}

  # some sort of error...
  myLog $w "rsync [expr {$r3 == 0 ? {threw system error} :\
                                        {returned completion code}}]: $r4\n"

  # pull the last N lines from the log and send them to 
  # PopError
  set N 7
  set last [string length $::glob(log)]
  while {[incr N -1] >= 0 && \
	     [set last [string last "\n" $::glob(log) ${last}-1]] > 0} {}
  PopError "[string range $::glob(log) ${last}+1 end]\
                  \nLast few output lines above. See Log for more."
}

